/*
 * naive bayesian classification algorithm implementation
 * 
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdint.h>
#include <sys/dir.h>
#include <sys/param.h>
#include <string.h>
#include <math.h>
#include <pthread.h>
#include <sched.h>
#include <errno.h>
#include <time.h>
#include <unistd.h>
#include <assert.h>
#include <sys/syscall.h>

#define MAX_PAYLOAD_LENGTH 1500

// must be the same as in sniff.c
#define PACKETS_TO_WAIT 4

#define noCOPY_NAME

typedef struct modelParamSingle {
     char protocolName[50];
     int protocolCode;
     int numSamples;
     int numFeatures;
     float sigma;
     float thres;
     int *Samples;
     int matched;
} MODEL_PARAMETERS_SINGLE;

int numModels;
int numFeatures;
MODEL_PARAMETERS_SINGLE *modelParametersSingle;
float *marginals;
int **firstComponentArray;

/* excluding tmp files */
int fileSelection(const struct direct *fname){
     if ((fname->d_name[0] != '.') && (fname->d_name[strlen(fname->d_name)-1] != '~'))
	  return 1;
     else
	  return 0;
}

/* load bayes models, returns the number of classes */
int readModelsSingle(char *nameDir) {
     int numFiles, i, j, h;
     struct direct **files;
     FILE *fp;
     char *completeFileName, *token;
     char string[100];

     /* find all files in the directory */
     numFiles = scandir(nameDir, &files, fileSelection, NULL);

     if (numFiles <= 0) {
	  fprintf(stderr, "No models found in the directory %s\n", nameDir);
	  exit(0);
     }

     // printf("There are %d SVM models.\n",numFiles);
     modelParametersSingle =
	  (MODEL_PARAMETERS_SINGLE *) malloc(numFiles * sizeof(MODEL_PARAMETERS_SINGLE));

     for(i = 0; i < numFiles; i++) {
	  completeFileName =
	       (char *) malloc(strlen(nameDir) +
			       strlen(files[i]->d_name) + 2);
	  strcpy(completeFileName, nameDir);
	  strcat(completeFileName, "/");
	  strcat(completeFileName, files[i]->d_name);

	  // load all parameters
	  if((fp = fopen(completeFileName, "r")) != NULL) {
	       while(fgets(string, sizeof(string), fp) != NULL) {
		    token = strtok(strtok(string, "\n"), "=");

		    if(strcmp(token, "protocolName") == 0)
			 strcpy(modelParametersSingle[i].protocolName, strtok(NULL, "="));
		    else if(strcmp(token, "protocolCode") == 0)
			 modelParametersSingle[i].protocolCode = atoi(strtok(NULL, "="));
		    else if(strcmp(token, "sigma") == 0)
			 modelParametersSingle[i].sigma = atof(strtok(NULL, "="));
		    else if(strcmp(token, "numFeatures") == 0)
			 modelParametersSingle[i].numFeatures = atoi(strtok(NULL, "="));
		    else if(strcmp(token, "thres") == 0)
			 modelParametersSingle[i].thres = atof(strtok(NULL, "="));
		    else if(strcmp(token, "numSamples") == 0)
			 modelParametersSingle[i].numSamples = atoi(strtok(NULL, "="));
		    else if(strcmp(token, "Samples") == 0) {
			 modelParametersSingle[i].Samples =
			      (int *) malloc(modelParametersSingle[i].numSamples *
					     modelParametersSingle[i].numFeatures *
					     sizeof(int));
			 *(modelParametersSingle[i].Samples) = atoi(strtok(strtok(NULL, "="), " "));
			 for(h = 1; h < modelParametersSingle[i].numFeatures; h++) 
			      *(modelParametersSingle[i].Samples + h) = atoi(strtok(NULL, " "));
			 for(j = 1; j < modelParametersSingle[i].numSamples; j++) {
			      if(fgets(string, sizeof(string), fp) == NULL) {
				   fprintf(stderr,
					   "Error in reading Samples: number of samples does "
					   "not correspond to the number of lines\n");
				   exit(-1);
			      }
					
			      *(modelParametersSingle[i].Samples +
				j * modelParametersSingle[i].numFeatures) =
				   atoi(strtok(strtok(string, "\n"), " "));
			      for(h = 1; h < modelParametersSingle[i].numFeatures; h++)
				   *(modelParametersSingle[i].Samples +
				     j * modelParametersSingle[i].numFeatures + h) =
					atoi(strtok(NULL, " "));
			 }
		    }
	       }
	       fclose(fp);
	  }
     }
     return numFiles;
}

/* initialize all data for the classifier */
int init(char* dir, int pktreceived) {
     int j, w, h;
     long i;

     numFeatures = pktreceived + 1;

     // load models
     numModels = readModelsSingle(dir);

     for(i = 0; i < numModels; i++) {
	  if(numFeatures > modelParametersSingle[i].numFeatures) {
               fprintf(stderr,
		       "Protocol model has not enough features.\n"
		       "Limit the number of packets to analyze or "
		       "provide more features in protocol model files\n");
               exit(0);
	  }
     }

     if(numModels == 0)
	  return -1;

     int max_i = numModels;
     int max_j = MAX_PAYLOAD_LENGTH;
     int max_h = numFeatures;
     int max_w, center, offset; 
     float sigma, coeff, coeffEsp;
     float *marg = (float *) malloc(max_j *sizeof(float));
     marginals = (float *) malloc(max_i*max_j*max_h *sizeof(float));
          
     // pre-compute marginals
     for(i = 0; i < max_i; i++){   // for each model
	  modelParametersSingle[i].matched = 0;    
	  max_w = modelParametersSingle[i].numSamples;
	  sigma = modelParametersSingle[i].sigma;
	  coeff = 1/(sqrt(2 * M_PI) * sigma * max_w);
	  coeffEsp = -1/(2 * pow(sigma, 2));

	  // pre-compute a marg
	  for(j = 0; j < max_j; j++){
               marg[j] = exp (coeffEsp * pow(j, 2));
	  }

	  for(h = 0; h < max_h; h++){   // for each dimension - feature
               offset = h * max_j + i * max_j * max_h;
               for(w = 0; w < max_w; w++){   // for each gaussiain
		    center = modelParametersSingle[i].Samples[max_h * w + h];
		    for(j = 0; j < max_j; j++){   // for each point of the marginal
			 marginals[j + offset] += coeff * marg[abs(j + 1 - center)];
		    }
               }
	  }
     }
     return 1;
}

/* 
 * check against a given model
 */
float compFunc(int model_num, int* testVectors) {
     float prod;
     int h;
     int proto_gap = model_num * numFeatures * MAX_PAYLOAD_LENGTH;

     for(prod = 1, h = 0; h < numFeatures; h++) {
	  prod *= marginals[testVectors[h] + MAX_PAYLOAD_LENGTH * h + proto_gap];
     }
     
     return prod;
}

/*
 * test flow classification, return the index of the class, -1 if unknown
 */
int testing(int* testVectors) {
     int candidateProto, j;
     float maxres, thres, res;
   
     for(maxres = -1, candidateProto = -1, j = 0;
	 j < numModels;
	 j++) {
	  res = compFunc(j, testVectors);
	  thres = (modelParametersSingle + j)->thres;

	  if(res > thres && res > maxres) {
	       maxres = res;
	       candidateProto = j;
	  }
     }
   
     return candidateProto;
}

/*
 * implementation of the classification algorithm, this function
 * receives the list of features and returns the application label
 */
int classify(int pktreceived, int *bytes_per_pkt, uint8_t proto,
	     uint16_t src_port, uint16_t dst_port, char *label) {
     int j;

     // one more to account for the protocol
     int features[PACKETS_TO_WAIT + 1];

     // exit if not enough packets: this must be checked by caller
     assert(pktreceived >= PACKETS_TO_WAIT);

     // fill the features vector up to PACKETS_TO_WAIT
     features[0] = proto - 1;
     for(j = 0; j < PACKETS_TO_WAIT; j++) {
	  features[j + 1] = bytes_per_pkt[j] - 1;
     }

     // classify and report result
     label[0] = 0;
     j = testing(features);
     if(j < 0) {
#ifdef COPY_NAME
	  strcpy(label, "unknown");
#endif
	  return j;
     }

#ifdef COPY_NAME
     strcpy(label, modelParametersSingle[j].protocolName);
#endif

     // must be thread safe
     __sync_fetch_and_add(&(modelParametersSingle[j].matched), 1);

     return modelParametersSingle[j].protocolCode;
}

void printStats() {
     int j;
     for(j = 0; j < numModels; j++) {
          fprintf(stderr,
		  "log: Model %d classified packet number: %d\n",
		  j, modelParametersSingle[j].matched);
     }
}
