/*
 * classification algorithm implementation
 * sigle-class svm
 * 
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/dir.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <string.h>
#include <math.h>
#include <pthread.h>
#include <sched.h>
#include <errno.h>
#include <time.h>
#include <unistd.h>
#include <sys/syscall.h>
#include <assert.h>
#include <stdint.h>

#define PACKETS_TO_WAIT 4

#define MAX_PAYLOAD_LENGTH 1500

typedef struct modelParamSingle {
     char protocolName[50];
     int protocolCode;
     int numSVs;
     int numFeatures;
     float rho;
     float sigma;
     float costanteEsp;
     float nu;
     float *alpha;
     int *SVs;
     int matched;
} MODEL_PARAMETERS_SINGLE;


int numModels;
MODEL_PARAMETERS_SINGLE *modelParametersSingle;
float *marginals;

/* excluding tmp files */
int fileSelection(const struct direct *fname){
     if((fname->d_name[0] != '.') && (fname->d_name[strlen(fname->d_name) - 1] != '~'))
	  return 1;
     else
	  return 0;
}

/* load svm models, returns the number of classes */
int readModelsSingle(char *nameDir) {
     int numFiles, i, j, h;
     struct direct **files;
     FILE *fp;
     char *completeFileName, *token;
     char string[100];

     /* find all files in the evaluated dir */
     numFiles = scandir(nameDir, &files, fileSelection, NULL);

     if (numFiles <= 0) {
	  fprintf(stderr, "No models found in the directory %s\n", nameDir);
	  exit(0);
     }

     // printf("There are %d SVM models.\n",numFiles);
     modelParametersSingle =
	  (MODEL_PARAMETERS_SINGLE *) malloc(numFiles * sizeof(MODEL_PARAMETERS_SINGLE));

     for (i = 0; i < numFiles; i++) {
	  completeFileName = (char *) malloc(strlen(nameDir) +
					     strlen(files[i]->d_name) + 2);
	  strcpy(completeFileName, nameDir);
	  strcat(completeFileName, "/");
	  strcat(completeFileName, files[i]->d_name);

	  // load all parameters
	  if((fp = fopen(completeFileName, "r")) != NULL) {
	       while(fgets(string, sizeof(string), fp) != NULL) {
		    token = strtok(strtok(string, "\n"), "=");

		    if(strcmp(token, "protocolName") == 0)
			 strcpy(modelParametersSingle[i].protocolName, strtok(NULL, "="));
		    else if(strcmp(token, "protocolCode") == 0)
			 modelParametersSingle[i].protocolCode = atoi(strtok(NULL, "="));
		    else if(strcmp(token, "rho") == 0)
			 modelParametersSingle[i].rho = atof(strtok(NULL, "="));
		    else if(strcmp(token, "sigma") == 0) {
			 modelParametersSingle[i].sigma = atof(strtok(NULL, "="));
			 modelParametersSingle[i].costanteEsp =
			      -1.0/(2.0 * ((modelParametersSingle+i)->sigma *
					   (modelParametersSingle+i)->sigma));
		    }
		    else if(strcmp(token, "nu") == 0)
			 modelParametersSingle[i].nu = atof(strtok(NULL, "="));
		    else if(strcmp(token, "numPkt") == 0)
			 modelParametersSingle[i].numFeatures = atoi(strtok(NULL, "="));
		    else if(strcmp(token, "numSVs") == 0)
			 modelParametersSingle[i].numSVs = atoi(strtok(NULL, "="));
		    else if(strcmp(token, "alpha") == 0) {
			 modelParametersSingle[i].alpha =
			      (float *) malloc(modelParametersSingle[i].numSVs *
					       sizeof(float));
			 *modelParametersSingle[i].alpha = atof(strtok(NULL, "="));
			 for(j = 1; j < modelParametersSingle[i].numSVs; j++) {
			      if(fgets(string, sizeof(string), fp) == NULL){
				   fprintf(stderr,
					   "Error in reading alphas: number of alphas does not "
					   "correspond to the number of lines\n");
				   exit(-1);
			      }
			      *(modelParametersSingle[i].alpha + j) = atof(strtok(string, "\n"));
			 }
		    } else if(strcmp(token, "SVs") == 0) {
			 modelParametersSingle[i].SVs =
			      (int *) malloc(modelParametersSingle[i].numSVs *
					     modelParametersSingle[i].numFeatures *
					     sizeof(int));
			 *(modelParametersSingle[i].SVs) = atoi(strtok(strtok(NULL, "="), " "));
			 for(h = 1; h < modelParametersSingle[i].numFeatures; h++)
			      *(modelParametersSingle[i].SVs + h) = atoi(strtok(NULL, " "));
			 for(j = 1; j < modelParametersSingle[i].numSVs; j++) {
			      if(fgets(string, sizeof(string), fp) == NULL) {
				   fprintf(stderr,
					   "Error in reading SVs: number of sv does not "
					   "correspond to the number of lines\n");
				   exit(-1);
			      }
			      *(modelParametersSingle[i].SVs +
				j * modelParametersSingle[i].numFeatures) =
				   atoi(strtok(strtok(string, "\n"), " "));
			      for(h = 1; h < modelParametersSingle[i].numFeatures; h++)
				   *(modelParametersSingle[i].SVs +
				     j * modelParametersSingle[i].numFeatures + h) =
					atoi(strtok(NULL, " "));
			 }
		    }
	       }
	       fclose(fp);
	  }
     }
     return numFiles;
}

/* initialize all data for the classifier */
int init(char* dir, int pktreceived) {
     int j;
     long i;
     float esp;

     // load models
     numModels = readModelsSingle(dir);

     for(i = 0; i < numModels; i++) {
	  if(pktreceived + 1 > modelParametersSingle[i].numFeatures) {
               fprintf(stderr,
		       "Protocol model has not enough features.\n"
		       "Limit the number of packets to analyze or "
		       "provide more features in protocol model files\n");
               exit(0);
	  }
     }

     //pre-compute marginals
     marginals = (float *) malloc(numModels * MAX_PAYLOAD_LENGTH *sizeof(float));
     for(i = 0; i < numModels; i++) {
	  modelParametersSingle[i].matched = 0;
	  for(j = 0; j < MAX_PAYLOAD_LENGTH; j++) {
               esp = -1/(2*pow(modelParametersSingle[i].sigma, 2));
               marginals[i * MAX_PAYLOAD_LENGTH + j] = exp(esp * pow(j, 2));
	  }
     }

     return 1;
}

/* 
 * check against a given model
 */
float compFunc(int model_num, int* testVectors) {
     int firstComponent, secondComponent, diff;
     float somma, prod;
     int h, w;
     int max_w = modelParametersSingle[model_num].numFeatures;
     int index = 0;

     somma = 0;

     for(h = 0; h < (modelParametersSingle + model_num)->numSVs; h++) {
          prod = 1;
          index = h * max_w;
          for(w = 0; w < max_w; w++) {
	       firstComponent = modelParametersSingle[model_num].SVs[w + index];
	       secondComponent = testVectors[w];
	       diff = firstComponent - secondComponent;
	       prod *= marginals[abs(diff) + model_num * MAX_PAYLOAD_LENGTH];
	  }
	  somma += (modelParametersSingle + model_num)->alpha[h] * prod;
     }

     return somma;
}

/*
 * test flow classification, return the index of the class, -1 if unknown
 */
int testing(int* testVectors){
     int candidateProto, j;
     float maxres, rho, res;

     for (maxres = -1, candidateProto = -1, j = 0;
	  j < numModels;
	  j++) {
	  res = compFunc(j, testVectors);
	  rho = (modelParametersSingle+j)->rho;

	  if(res > rho && res > maxres) {
	       maxres = res;
	       candidateProto = j;
	  }
     }
     return candidateProto;
}

/*
 * implementation of the classification algorithm, this function
 * receives the list of features and returns the application label
 */
int classify(int pktreceived, int *bytes_per_pkt, uint8_t proto,
	     uint16_t src_port, uint16_t dst_port, char *label) {
     int j;

     // one more to account for the protocol
     int features[PACKETS_TO_WAIT + 1];

     // exit if not enough packets: this must be checked by caller
     assert(pktreceived >= PACKETS_TO_WAIT);

     // fill the features vector up to PACKETS_TO_WAIT
     features[0] = proto;
     for(j = 0; j < pktreceived; j++)
	  features[j + 1] = bytes_per_pkt[j];

     // classify and report result
     label[0] = 0;
     j = testing(features);
     if(j < 0) {
#ifdef COPY_NAME
	  strcpy(label, "unknown");
#endif
	  return j;
     }

#ifdef COPY_NAME
     strcpy(label, modelParametersSingle[j].protocolName);
#endif

     // must be thread safe
     __sync_fetch_and_add(&(modelParametersSingle[j].matched), 1);

     return modelParametersSingle[j].protocolCode;
}

void printStats(void) {
     int j;
     for(j = 0; j < numModels; j++) {
          fprintf(stderr, "log: Model %d classified packet number: %d\n",
		  j, modelParametersSingle[j].matched);
     }
}
