#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <sys/socket.h>
#include <netpacket/packet.h>
#include <net/ethernet.h>
#include <signal.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <string.h>
#include <net/if.h>
#include <netinet/ether.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <dlfcn.h>
#include <unistd.h> 
#include <sys/wait.h>
#include <assert.h>
#include <pcap.h>

// pay attention, ip checksum is not changed!!
#define noCHANGE_TOS

#define noDOUBLE_HASH
#define noDEBUG
#define noLOG_FLOW_RUNTIME
#define noLOG_FLOW_ENDTIME
#define LOG_FLOW        60

#define CLASS_THREAD_NUMBER 6

#define TERMINATE_DELAY     1000000

#define PACKET_MAX_LENGTH      9500
#define MAX_FLOW_AGE_SEC         15
#define PACKETS_TO_WAIT           4
#define ETHERTYPE_MUX_IN     0x9999
#define ETHERTYPE_MUX_OUT    0x9998
#define ETHERTYPE_IP         0x0800
#define JOB_RING_SIZE          1000
#define MAX_IFACE                 5
#define ETHERNET_HEADER_LENGTH   14
#define ETHERTYPE_OFFSET         12
#define IP_HEADER_LENGTH         20
#define TCP_HEADER_LENGTH_SHORT   4
#define IP_BYTE0_OFFSET           0
#define IP_TOS_OFFSET             1
#define IP_TOTAL_LENGTH_OFFSET    2

typedef unsigned char u8;
typedef unsigned int u32;
typedef unsigned short u16;

#ifdef DEBUG
#define LOG_DEBUG(format, ...)			\
     fprintf(stdout, "debug: " format ,		\
	     ## __VA_ARGS__ )
#else
#define LOG_DEBUG(format, ...)       {}
#endif

#define LOG(format, ...)			\
     fprintf(stdout, "log: " format ,		\
	     ## __VA_ARGS__ )

#define LOG_ERROR(format, ...)			\
     fprintf(stderr, "error: " format ,		\
	     ## __VA_ARGS__ )

/* mux header. It precedes every packet in the file. */
typedef struct muxhdr {
     u16 caplen;                 /* length of portion present */
     u16 pktlen;                 /* length of this packet (on the wire) */
     u16 islast;
     u8 pad[2];
} muxhdr_t;

typedef struct _flow {
     u32 ipsrc;
     u32 ipdst;
     u16 portsrc;
     u16 portdst;
     u8 proto;
     int timeLastSeen;
     int empty;
     sig_atomic_t seenPackets;
     int packet_sizes[PACKETS_TO_WAIT];
     sig_atomic_t classificationCode;
     sig_atomic_t readyForClassification;
     sig_atomic_t classified;
     sig_atomic_t lock;
     struct _flow* next;
} flow_t;

typedef struct _job {
     sig_atomic_t busy;
     flow_t* flow;
     int pr_done;
     int pr_taken;
} job_t;

typedef struct _flow_ptr {
     flow_t* first;
     flow_t* last;
     sig_atomic_t locked;
} flow_ptr_t;

/* "Classifier" data structure associated to each loaded classification algorithm */
typedef struct _classifier {
     char name[100];             /* name of the classification algorithm */
     void *handle;               /* file handle returned during the classifier loading */

     int (*init_fn)(char*, int); /* classifier's initialization function,
				  * receives the directory containing the classifiers and
				  * the number of pkts */

     /* classification function, receives the number of pkts, the packet size array, 
	the transport protocol, the src and dst ports, the pointer to the payload bytes 
	and the pointer to the application label to fill. Return a code number 
	associated to each protocol */
     int (*classify_fn)(int, int*, u8, u16, u16, char*);

     /* statistics function, called at the end to display useful information */
     void (*stats_fn)(void);
} classifier_t;

struct _forwarder;

typedef struct _forwarder_thread {
     struct _forwarder* fwdData;
     pthread_t thread;
     int ifaceIndex;
     sig_atomic_t receivedPackets;
     sig_atomic_t inputErrors;
     sig_atomic_t outputErrors;
     sig_atomic_t outputFailed;
     sig_atomic_t totAllocFlows;
     sig_atomic_t oldFlows;
     volatile sig_atomic_t running;
     volatile sig_atomic_t stopped;
} forwarder_thread_t;

typedef struct _classifier_thread {
     struct _forwarder* fwdData;
     pthread_t thread;
     int id;
     volatile sig_atomic_t classifying;
     volatile sig_atomic_t stopped;
} classifier_thread_t;

typedef struct _forwarder {
     char* interfaceName[MAX_IFACE];
     char* traceName;
     pthread_t* timeThread;
     forwarder_thread_t forwarderThread[MAX_IFACE];
     forwarder_thread_t pcapThread;
     classifier_thread_t classifierThread[CLASS_THREAD_NUMBER];
     int inputHandler[MAX_IFACE];
     int outputHandler[MAX_IFACE];
     int ifaceNumber, ifaceNumberInitialized;
     struct sockaddr_ll outputLL[MAX_IFACE];
     flow_ptr_t* flowTable;
     sig_atomic_t startTimeSec;
     sig_atomic_t currentTimeSec;
     sig_atomic_t lastPacketTimeSec;
     struct timeval interruptTime;

     job_t* jobRing;
     volatile sig_atomic_t nextJobToStore;
     volatile sig_atomic_t nextJobToProcess;
     volatile sig_atomic_t job_busy;
     sig_atomic_t processedJobs;

     //variables for statistics
     sig_atomic_t classifiedPackets;
     sig_atomic_t nonIpPackets;
     sig_atomic_t nonIp45Packets;
     sig_atomic_t processedPackets;
     sig_atomic_t notSupportedPackets;
     sig_atomic_t notClassifiedPackets;
     sig_atomic_t flowNumber;

     classifier_t* classifier;
#ifdef LOG_FLOW
     int classified_flows;
     sig_atomic_t flow_whose_Nth_packet_was_marked[LOG_FLOW];
     sig_atomic_t flow_classified_exactly_after_N_packets[LOG_FLOW];
     sig_atomic_t flow_with_at_least_N_packets[LOG_FLOW];
#endif // LOG_FLOW

     pcap_t* pcapReader;
     char pcapErrBuf[PCAP_ERRBUF_SIZE];
     int pcapOffset;

} forwarder_t;

#ifdef DOUBLE_HASH
#define FLOW_TABLE_SIZE_1 199033   /* Prime number */
#define FLOW_TABLE_SIZE_2 397
#define FLOW_TABLE_SIZE (FLOW_TABLE_SIZE_1*FLOW_TABLE_SIZE_2) /* 5000000 */
#else
#define FLOW_TABLE_SIZE           5000000
#define FLOW_TABLE_SIZE_1 FLOW_TABLE_SIZE
#define FLOW_TABLE_SIZE_2 FLOW_TABLE_SIZE
#endif

inline unsigned int flowHash(u8* ipPayload, int ipHLen) {
     u32* ip2Hash = (u32*) (ipPayload + 12);
     u16* tcp2Hash = (u16*) (ipPayload + ipHLen);
     u8* proto = ipPayload + 9;
     u32 ipsrc = *ip2Hash;
     u32 ipdst = *(ip2Hash + 1);
     u32 portsrc = (u32) (*tcp2Hash);
     u32 portdst = (u32) (*(tcp2Hash + 1));
     u32 ipsrc_tmp = ipsrc * 59;

     unsigned int hash = (
	  (ipsrc_tmp)     ^
          (ipdst)         ^
          (portsrc << 16) ^
          (portdst << 0)  ^
          *proto) % FLOW_TABLE_SIZE_1;
     
#ifdef DOUBLE_HASH
     unsigned int hash2 = (
	  (ipsrc_tmp)     +
          (ipdst)         +
          (portsrc << 16) +
          (portdst << 0)) % FLOW_TABLE_SIZE_2;
     hash = hash * FLOW_TABLE_SIZE_2 + hash2;
#endif

     return hash;
}

void terminate_handler(int datas, siginfo_t* info, void* nu) {
     struct sigaction saio;    
     saio.sa_sigaction = terminate_handler;
     sigemptyset(&saio.sa_mask);
     saio.sa_flags |= SA_SIGINFO;
     saio.sa_flags &= ~SA_RESTART;
     if( sigaction(SIGINT, &saio, NULL) == -1) {
	  LOG_ERROR("Unrecoverable error, terminating\n");
	  exit(-1);
     }
}

/* set empty all flows in table and cut the list */
int initFlowTable(forwarder_t* fwdData, int flowTableSize) {
     int i;
     flow_t* flow1StLevel = malloc(flowTableSize * sizeof(flow_t));
     flow_ptr_t* ptr = malloc(flowTableSize * sizeof(flow_ptr_t));
     if(!ptr || !flow1StLevel)
	  return 0;
     for(i = 0; i < flowTableSize; i++) {
	  ptr[i].first = ptr[i].last = &(flow1StLevel[i]);
	  ptr[i].locked = 0;
	  flow1StLevel[i].empty = 1;
	  flow1StLevel[i].next = NULL;
     }
     fwdData->flowTable = ptr;
     return flowTableSize;
}

inline int flowCompare(flow_t* flow, u8* ipPayload, int ipHLen) {
     u32* ip2Compare = (u32*) (ipPayload + 12);
     u16* tcp2Compare = (u16*) (ipPayload + ipHLen);
     u8* proto = ipPayload + 9;

     if(flow->ipsrc != *ip2Compare ||
	flow->ipdst != *(ip2Compare + 1) ||
	flow->portsrc != *tcp2Compare ||
	flow->portdst != *(tcp2Compare + 1) ||
	flow->proto != *proto)
	  return 1;
     return 0;
}

void flowInsert(flow_t* flow, u8* ipPayload, int ipHLen, int currentTimeSec) {
     u32* ip2Insert = (u32*) (ipPayload + 12);
     u16* tcp2Insert = (u16*) (ipPayload + ipHLen);
     u8* proto = ipPayload + 9;
     
     flow->ipsrc = *ip2Insert;
     flow->ipdst = *(ip2Insert + 1);
     flow->portsrc = *tcp2Insert;
     flow->portdst = *(tcp2Insert + 1);
     flow->proto = *proto;
     flow->timeLastSeen = currentTimeSec;
     flow->empty = 0;
     flow->seenPackets = 0;
     flow->readyForClassification = 0;
     flow->classified = 0;
     flow->classificationCode = -1;
     flow->lock = 0;
}

// data points to first byte of ip header
int processIpPacketData(forwarder_thread_t* fwdThread, u8* data, int length) {
     int hash;
     flow_ptr_t* flowList;
     flow_t** curFlow;
     forwarder_t* fwdData = fwdThread->fwdData;     

     __sync_fetch_and_add(&(fwdData->processedPackets), 1);

     if(length < IP_HEADER_LENGTH + TCP_HEADER_LENGTH_SHORT ||
	data[0] != 0x45) {
	  __sync_fetch_and_add(&(fwdData->nonIp45Packets), 1);
	  return -1;
     }

     hash = flowHash(data, IP_HEADER_LENGTH);

     // pay attention when you decide (if you do this!) to remove this lock,
     // read comment XXX1 below
     while(__sync_lock_test_and_set(&(fwdData->flowTable[hash].locked), 1));

     flowList = &(fwdData->flowTable[hash]);
     curFlow = &flowList->first;

     // rules
     // 1. empty flows stay at the end of the list
     // 2. walk the list, for each flow in the list
     //    i) if an empty flow is found, exit;
     //    ii) if the flow is found, exit;
     //    iii) if an old flow is found, move to the end of the list and mark it as empty
     //         pay attention, if it is the last flow in the list, just mark it as empty and exit
     //    iv) if the list is finished, add a new empty node at the end and exit

     while(1) {
	  if( (*curFlow)->empty == 1) {
	       LOG_DEBUG("Found empty slot\n");
	       break;
	  }

	  if(!flowCompare(*curFlow, data, IP_HEADER_LENGTH)) {
	       LOG_DEBUG("Found flow\n");
	       break;
	  }
	  
	  // XXX1 we don't care about locking to check readyForClassification because
	  // we lock the entire line of the hash table, another thread will wait
	  if((*curFlow)->readyForClassification == 0 && fwdData->currentTimeSec - (*curFlow)->timeLastSeen > MAX_FLOW_AGE_SEC) {
	       if(*curFlow == flowList->last) {
		    LOG_DEBUG("Found old flow at end of the list\n");
		    (*curFlow)->empty = 1;
		    break;
	       } else {
		    LOG_DEBUG("Found old flow in list, moving to the end\n");
		    flow_t* tmp = (*curFlow)->next;
		    (*curFlow)->empty = 1;
		    (*curFlow)->next = NULL;
		    flowList->last->next = *curFlow;
		    flowList->last = *curFlow;
		    *curFlow = tmp;
		    fwdThread->oldFlows++;
		    continue;
	       }
	  }

	  if((*curFlow)->next == NULL) {
	       flow_t* tmp = malloc(sizeof(flow_t));
	       if(!tmp) {
		    LOG("out of memory for this flow, skipping\n");
		    goto exit_flow;
	       }
	       LOG_DEBUG("Inserting new flow at the end of the list\n");
	       tmp->empty = 1;
	       tmp->next = NULL;
	       flowList->last = tmp;
	       (*curFlow)->next = tmp;
	       curFlow = &((*curFlow)->next);
	       fwdThread->totAllocFlows++;
	       break;
	  }

	  LOG_DEBUG("Moving to next element\n");
	  curFlow = &( (*curFlow)->next );
     }

     if( (*curFlow)->empty == 1 ) {
	  LOG_DEBUG("Inserting new flow into list\n");
	  __sync_fetch_and_add(&(fwdData->flowNumber), 1);
	  flowInsert(*curFlow, data, IP_HEADER_LENGTH, fwdData->currentTimeSec);
     } else {
	  LOG_DEBUG("Refreshing flow age\n");
	  (*curFlow)->timeLastSeen = fwdData->currentTimeSec;
     }

     if((*curFlow)->seenPackets < PACKETS_TO_WAIT) {
	  u16* totalLength = (u16*) &(data[IP_TOTAL_LENGTH_OFFSET]);
	  (*curFlow)->packet_sizes[(*curFlow)->seenPackets] = (int) ntohs(*totalLength);
     }
#ifdef LOG_FLOW
     __sync_fetch_and_add(&((*curFlow)->seenPackets), 1);
#else
     (*curFlow)->seenPackets++;
#endif

     if(fwdData->classifier != NULL) {

	  // packet from pcap
	  if(fwdData->pcapReader != NULL) {
	       if((*curFlow)->seenPackets >= PACKETS_TO_WAIT &&
		  (*curFlow)->classified == 0) {
		    char classificationOutput[30] = "";
		    (*curFlow)->classificationCode = 
			 fwdData->classifier->classify_fn(PACKETS_TO_WAIT, 
						     (*curFlow)->packet_sizes, 
						     (*curFlow)->proto, 
						     ntohs((*curFlow)->portsrc), 
						     ntohs((*curFlow)->portdst), 
						     classificationOutput);
		    (*curFlow)->classified = 1;
	       }
	  }
	  // packet from (some) interface
	  else {
	       // lock the flow to avoid race condition
	       // given the same reason XXX1 we are the only (input) thread that can access
	       // this flow. Maybe the classifier thread can.
	       // XXX2
	       if( (*curFlow)->seenPackets >= PACKETS_TO_WAIT &&
		   (*curFlow)->classified == 0 &&
		   (*curFlow)->readyForClassification == 0) {
		    while(__sync_lock_test_and_set(&(fwdData->job_busy), 1));
		    if( (fwdData->nextJobToStore + 1) % JOB_RING_SIZE != fwdData->nextJobToProcess ) {
			 (*curFlow)->readyForClassification = 1;
			 fwdData->jobRing[fwdData->nextJobToStore].flow = (*curFlow);
			 fwdData->jobRing[fwdData->nextJobToStore].pr_done = 0;
			 fwdData->jobRing[fwdData->nextJobToStore].pr_taken = 0;
			 fwdData->nextJobToStore = (fwdData->nextJobToStore + 1) % JOB_RING_SIZE;
		    } else {
			 LOG_DEBUG("processPacket: job ring was full\n");
		    }
		    __sync_lock_release(&(fwdData->job_busy));
	       }
	  }
     }

exit_flow:
     //insert in the packet the classification code
     if (fwdData->classifier != NULL) {

#ifdef LOG_FLOW
	  if((*curFlow)->seenPackets < LOG_FLOW)
	       __sync_fetch_and_add(&(fwdData->flow_with_at_least_N_packets[(*curFlow)->seenPackets]), 1);
#endif // LOG_FLOW

	  if ((*curFlow)->classified == 1) {

#ifdef LOG_FLOW
	       if((*curFlow)->seenPackets < LOG_FLOW)
		    __sync_fetch_and_add(&(fwdData->flow_whose_Nth_packet_was_marked[(*curFlow)->seenPackets]), 1);
#endif // LOG_FLOW

#ifdef CHANGE_TOS
	       data[IP_TOS_OFFSET] = (char)((*curFlow)->classificationCode) & 0xFF;
#endif

	       __sync_fetch_and_add(&(fwdData->classifiedPackets), 1);
	  } else if ((*curFlow)->classified == 0 && (*curFlow)->seenPackets <= PACKETS_TO_WAIT){
	       __sync_fetch_and_add(&(fwdData->notSupportedPackets),1);
	  } else{
	       __sync_fetch_and_add(&(fwdData->notClassifiedPackets),1);
	  }
     }

     __sync_lock_release(&(fwdData->flowTable[hash].locked));
    
     return 0;
}

// data points to the l3 payload
// length is the length of the l3 payload
// eth_type points to the ethertype of the incoming packet
int checkMuxAndProcessPacket(forwarder_thread_t* fwdThread,
			     u8* data,
			     int length,
			     u16* eth_type) {

     if(ntohs(*eth_type) == ETHERTYPE_MUX_IN) {

	  // mux case: assume each muxed packet has ethernet header
	  // and no fcs at end
	  int remaining_bytes = length;

	  while(remaining_bytes) {
	       muxhdr_t* muxhdr = (muxhdr_t*) data;
	       int islast = muxhdr->islast;
	       int caplen = ntohs(muxhdr->caplen);
	       data += sizeof(muxhdr_t);
	       if(caplen == 0) {
		    LOG_ERROR("Found malformed packet inside mux, skipping jumbo frame\n");
		    break;
	       }
	       else if(caplen > ETHERNET_HEADER_LENGTH &&
		       ntohs(*(u16*)(data + ETHERNET_HEADER_LENGTH - 2)) == ETHERTYPE_IP) {
		    processIpPacketData(fwdThread,
				      data + ETHERNET_HEADER_LENGTH,
				      caplen - ETHERNET_HEADER_LENGTH);
	       } else {
		    __sync_fetch_and_add(&(fwdThread->fwdData->nonIpPackets), 1);
	       }
	       remaining_bytes -= sizeof(muxhdr_t) + caplen;
	       data += caplen;
	       if(islast == 1) break;
	  }
     }
     else if(ntohs(*eth_type) == ETHERTYPE_IP) {
	  // non-mux case and this is IP
	  processIpPacketData(fwdThread, data, length);
     }
     else {
	  forwarder_t* fwdData = fwdThread->fwdData;
	  __sync_fetch_and_add(&(fwdData->nonIpPackets), 1);
     }
     return 0;
}

int initClassifier(forwarder_t* fwdData, int jobRingSize, char *classifierName, char *modelsDir) {
     classifier_t* ptr = NULL;

     if(!classifierName) {
	  LOG("Classifier not loaded\n");
	  goto class_exit_done;
     }

     ptr = (classifier_t*)malloc(sizeof(classifier_t));

     if (!ptr) {
	  LOG_ERROR("Out of memory: cannot create classifier\n");
	  goto class_exit_err;
     }
     ptr->handle = dlopen(classifierName, RTLD_NOW);
     
     if (!ptr->handle) {
	  LOG_ERROR("Cannot load classifier %s\n", classifierName);
	  goto class_exit_err;
     }
             
     ptr->classify_fn = dlsym(ptr->handle, "classify");
     ptr->stats_fn = dlsym(ptr->handle, "printStats");
     ptr->init_fn = dlsym(ptr->handle, "init");
     ptr->init_fn(modelsDir, PACKETS_TO_WAIT);
     LOG("Classifier %s loaded\n", classifierName);

     fwdData->jobRing = malloc(jobRingSize * sizeof(job_t));
     if(!fwdData->jobRing) {
	  LOG_ERROR("Cannot allocate job ring\n");
	  goto class_exit_err;
     }

     fwdData->job_busy = 0;
     fwdData->nextJobToProcess = 0;
     fwdData->nextJobToStore = 0;
   
class_exit_done:
     fwdData->classifier = ptr;
     return jobRingSize;

class_exit_err:
     LOG_ERROR("Classifier %s (with models in %s) initialization failed\n",
	       classifierName, modelsDir);
     return 0;
}

int initForwarder(forwarder_t* fwdData) {
     struct ifreq ifr, oifr;
     struct sockaddr_ll addr;
     char* interfaceName = NULL;
     char* outputInterfaceName = NULL;
     char* inputInterfaceName = NULL;
     int isoc, osoc;
     int i;

     for(i = 0; i < fwdData->ifaceNumber; i++) {

	  fwdData->inputHandler[i] = fwdData->outputHandler[i] = -1;
	  fwdData->ifaceNumberInitialized++;

          interfaceName = fwdData->interfaceName[i];

          // check if output interface is given after input, e.g., eth0,eth1
          // this code breaks interfaceName into first part (input) and
          // a second string starting with the output interface specification
          outputInterfaceName = strchr(interfaceName, ',');
          if(outputInterfaceName != NULL) {
               *outputInterfaceName = '\0';
               outputInterfaceName++;
          }
          inputInterfaceName = interfaceName;

          isoc = socket(PF_PACKET, SOCK_RAW, htons(ETH_P_ALL));
          if(isoc < 0) {
	       LOG_ERROR("Can't create raw socket for interface %s (input)\n", inputInterfaceName);
	       goto fwdinit_exit_err;
          }
          fwdData->inputHandler[i] = isoc;

          memset(&ifr, 0, sizeof(ifr));
          strncpy(ifr.ifr_name, inputInterfaceName, sizeof(ifr.ifr_name)-1);
          if(ioctl(isoc, SIOCGIFINDEX, &ifr) == -1) {
	       LOG_ERROR("Invalid interface name %s (input)\n", inputInterfaceName);
	       goto fwdinit_exit_err;
          }
	  LOG("Interface %s is at index %d\n", inputInterfaceName, ifr.ifr_ifindex);
          memset(&addr, 0, sizeof(addr));
          addr.sll_family = AF_PACKET;
          addr.sll_protocol = htons(ETHERTYPE_MUX_IN);
          addr.sll_ifindex = ifr.ifr_ifindex;
          if(bind(isoc, (struct sockaddr*) &addr, sizeof(addr)) == -1) {
	       LOG_ERROR("Can't bind raw socket to interface %s: %s\n", inputInterfaceName, strerror(errno));
	       goto fwdinit_exit_err;
          }

          if(outputInterfaceName != NULL) {

               osoc = socket(PF_PACKET, SOCK_RAW, htons(ETH_P_ALL));
               if(osoc < 0) {
		    LOG_ERROR("Can't create raw socket for interface %s (output)\n", outputInterfaceName);
                    goto fwdinit_exit_err;
               }
	       fwdData->outputHandler[i] = osoc;

               memset(&oifr, 0, sizeof(oifr));
               strncpy(oifr.ifr_name, outputInterfaceName, sizeof(oifr.ifr_name)-1);
               if(ioctl(osoc, SIOCGIFINDEX, &oifr) == -1) {
		    LOG_ERROR("Invalid interface name %s (output)\n", inputInterfaceName);
                    goto fwdinit_exit_err;
               }
               fwdData->outputLL[i].sll_family = PF_PACKET;
               fwdData->outputLL[i].sll_halen = 6;
               fwdData->outputLL[i].sll_ifindex = oifr.ifr_ifindex;
	       LOG("Output interface is at index %d\n", oifr.ifr_ifindex);
          }
     }

     return fwdData->ifaceNumber;

fwdinit_exit_err:
     LOG_ERROR("Forwarding subsystem initialization failed\n");
     return 0;
}

int initPcapReader(forwarder_t* fwdData) {
     int datalink;

     fwdData->pcapReader = pcap_open_offline(fwdData->traceName, fwdData->pcapErrBuf);
     if(fwdData->pcapReader == NULL) {
	  LOG_ERROR("Cannot open pcap for reading: %s\n", fwdData->pcapErrBuf);
	  goto pcapinit_exit_err;
     }
     datalink = pcap_datalink(fwdData->pcapReader);

     switch(datalink) {
     case DLT_C_HDLC:
          fwdData->pcapOffset = 2;
	  break;
     case DLT_EN10MB:
	  fwdData->pcapOffset = 12;
	  break;
     default:
	  LOG_ERROR("Unsopported datalink type in pcap file\n");
	  goto pcapinit_exit_err;
     }
     return 1;

pcapinit_exit_err:
     LOG_ERROR("Pcap Reader initialization failed\n");
     return 0;
}

void* timerFunc(void* ptr) {
     forwarder_t* fwdData = (forwarder_t*) ptr;
     struct timeval tv;
     while(1) {
	  gettimeofday(&tv, NULL);
	  fwdData->currentTimeSec = tv.tv_sec;
	  sleep(1);
     }
}

void* pcapFunc(void* ptr) {
     forwarder_thread_t* pcapThread = (forwarder_thread_t*) ptr;
     // int ifaceIndex = (int) (pcapThread->ifaceIndex);
     forwarder_t* fwdData = pcapThread->fwdData;

     struct pcap_pkthdr* pkthdr;
     u8* buffer = NULL;
     int retc = -1;

     while(pcapThread->running) {
	  retc = pcap_next_ex(fwdData->pcapReader, &pkthdr, (const u_char**) &buffer);
          if(retc == -1) {
               LOG_ERROR("Error reading pcap, fatal error: %s\n", pcap_geterr(fwdData->pcapReader));
               break;
          }
          else if( retc == -2 ) {
               LOG("Trace ended\n");
               break;
          }

	  pcapThread->receivedPackets++;

	  // remove two bytes of the ethertype field
	  checkMuxAndProcessPacket(pcapThread,
				   buffer + fwdData->pcapOffset + 2,
				   pkthdr->caplen - fwdData->pcapOffset - 2,
				   (u16*)(buffer + fwdData->pcapOffset));
     }

     pcapThread->running = 0;
     pcapThread->stopped = 1;
     kill(getpid(), SIGINT);

     return NULL;
}

void* forwardFunc(void* ptr) {
     forwarder_thread_t* fwdThread = (forwarder_thread_t*) ptr;
     int ifaceIndex = (int) (fwdThread->ifaceIndex);
     forwarder_t* fwdData = fwdThread->fwdData;
     struct sockaddr_ll from;
     socklen_t fromlen = sizeof(struct sockaddr_ll);
     u8* buffer = (u8*) malloc(PACKET_MAX_LENGTH * sizeof(u8));
     u16* type;
     int retc, retc_send;

     while(fwdThread->running) {
	  retc = recvfrom(fwdData->inputHandler[ifaceIndex],
			  buffer,
			  PACKET_MAX_LENGTH,
			  0, //MSG_DONTWAIT,
			  (struct sockaddr*)&from, &fromlen);
	  
	  if(retc == -1) {
	       switch(errno) {
	       case EINTR:
		    LOG_DEBUG("recvfrom returned -1 + EINTR\n");
		    continue;

	       case EAGAIN:
		    LOG_DEBUG("recvfrom returned -1 + EAGAIN\n");
		    continue;

	       default:
		    LOG_ERROR("recvfrom returned -1 with errors: %s", strerror(errno));
		    fwdThread->inputErrors++;
		    // sleep(1);
		    continue;
	       }
	  }

	  fwdData->lastPacketTimeSec = fwdData->currentTimeSec;

	  if(from.sll_pkttype == PACKET_OUTGOING) {
	       LOG("forwardFunc: ignoring outgoing packet\n");
	       continue;
	  } 

	  fwdThread->receivedPackets++;

	  type = (u16*)(buffer + ETHERTYPE_OFFSET);
	  checkMuxAndProcessPacket(fwdThread,
				   buffer + ETHERNET_HEADER_LENGTH,
				   retc - ETHERNET_HEADER_LENGTH,
				   type);

	  *type = htons(ETHERTYPE_MUX_OUT);

	  if(fwdData->outputHandler[ifaceIndex] != -1) {
	       retc_send = sendto(fwdData->outputHandler[ifaceIndex],
				  buffer,
				  retc,
				  0,
				  (struct sockaddr*)(&from), 
				  sizeof(struct sockaddr_ll));
	       
	       //if(retc_send == -1) {
	       if(retc_send == -1) {
		    LOG_DEBUG("Error sending packet from interface %d\n", ifaceIndex);
		    fwdThread->outputErrors++;
	       }
	       else if(retc_send < -1)
	       {
		    fwdThread->outputFailed++;
	       }
	  }
     }
     fwdThread->stopped = 1;
     return NULL;
}

sig_atomic_t job_busy_new;
int getJobFromRing(classifier_thread_t* clsThread) {
     forwarder_t* fwdData = clsThread->fwdData;
     while(__sync_lock_test_and_set(&job_busy_new, 1));
     while(clsThread->classifying) {
          if(fwdData->nextJobToProcess != fwdData->nextJobToStore) {
               // ok we have chance that a new job is available, now try to
               // free leftmost packets in the ring, if any
               int index;
               if(fwdData->jobRing[fwdData->nextJobToProcess].pr_done > 0) {
                    fwdData->nextJobToProcess = (fwdData->nextJobToProcess + 1) % JOB_RING_SIZE;
                    continue;
               } else {
                    index = fwdData->nextJobToProcess;
                    while(clsThread->classifying) {
                         if(index != fwdData->nextJobToStore) {
                              if(fwdData->jobRing[index].pr_taken == 0) {
                                   fwdData->jobRing[index].pr_taken = 1;
                                   __sync_lock_release(&job_busy_new);
                                   return index;
                              }
                              index = (index + 1) % JOB_RING_SIZE;
                         }
                    }
               }
          }
     }
     __sync_lock_release(&job_busy_new);
     return -1;
}

void* classFunc(void* ptr) {
     classifier_thread_t* clsThread = (classifier_thread_t*) ptr;
     // int id = (int) (clsThread->id);
     forwarder_t* fwdData = clsThread->fwdData;
     char classificationOutput[30] = "";
   
     while(clsThread->classifying) {
	  job_t* job;
	  flow_t* flow;
	  int index = getJobFromRing(clsThread);

	  if(index == -1)
	       break;

	  job = &fwdData->jobRing[index];
	  flow = job->flow;

	  // We lock the flow to avoid race conditions (if any)
	  // Locking should work also as memory barrier
	  while(__sync_lock_test_and_set(&(flow->lock), 1));

	  //classification
	  if (fwdData->classifier != NULL){
	       flow->classificationCode = 
		    fwdData->classifier->classify_fn(PACKETS_TO_WAIT, 
						     flow->packet_sizes, 
						     flow->proto, 
						     ntohs(flow->portsrc), 
						     ntohs(flow->portdst), 
						     classificationOutput);
	  }
#ifdef LOG_FLOW_RUNTIME
	  struct in_addr* ipsrc = (struct in_addr*) &(flow->ipsrc);
	  struct in_addr* ipdst = (struct in_addr*) &(flow->ipdst);
	  
	  LOG("%s:%hu \t-> \t", inet_ntoa(*ipsrc), ntohs(flow->portsrc));
	  LOG("%s:%hu \t", inet_ntoa(*ipdst), ntohs(flow->portdst));
	  LOG("proto(%hhu) \t", flow->proto);
	  LOG("[%d %d %d %d] \t",
	      flow->packet_sizes[0],
	      flow->packet_sizes[1],
	      flow->packet_sizes[2],
	      flow->packet_sizes[3]);
	  if (fwdData->classifier != NULL)
	       LOG("%s\n", classificationOutput);  
	  else
	       LOG("\n");
#endif // LOG_FLOW_RUNTIME
 
	  // release this job and point to the next
	  flow->classified = 1;
#ifdef LOG_FLOW
	  __sync_fetch_and_add(&(fwdData->classified_flows),1);
	  if(flow->seenPackets < LOG_FLOW) {
	       __sync_fetch_and_add(&(fwdData->flow_classified_exactly_after_N_packets[flow->seenPackets]), 1);
	  }
#endif // LOG_FLOW
	  __sync_lock_release(&(flow->lock));
	  // we set this AFTER the memory barrier so that we are sure
	  // that the two checks in XXX2 will not fail.
	  flow->readyForClassification = 0;
	  job->pr_done = 1;
     }
     clsThread->stopped = 1;
     return NULL;
}

int initMainLoop(forwarder_t* fwdData) {

     struct sigaction saio;
     struct sched_param mysched;

     memset(&saio, 0, sizeof(struct sigaction));
     saio.sa_sigaction = terminate_handler;
     sigemptyset(&saio.sa_mask);
     saio.sa_flags |= SA_SIGINFO;
     saio.sa_flags &= ~SA_RESTART;
     if(sigaction(SIGINT, &saio, NULL) == -1) {
	  LOG_ERROR("Error installing CTRL+C handler: %s\n", strerror(errno));
          goto proc_exit_err;
     }

     // Setpriority to RealTime and increase nice level
     setpriority(PRIO_PROCESS, 0, -10);
     memset(&mysched, 0, sizeof(struct sched_param));
     mysched.sched_priority = 2;
     if(sched_setscheduler(0, SCHED_RR, &mysched) == -1) {
	  LOG_ERROR("Can't set maximum priority\n");
	  goto proc_exit_err;
     }

     gettimeofday(&(fwdData->interruptTime), NULL);
     fwdData->startTimeSec = 
	  fwdData->currentTimeSec = 
	  fwdData->interruptTime.tv_sec;

     return 1;

proc_exit_err:
     return 0;
}

int initThread(forwarder_t* fwdData) {
     int i;
     int priorityMin = sched_get_priority_min(sched_getscheduler(getpid()));
     int priorityMax = sched_get_priority_max(sched_getscheduler(getpid()));

     for(i = 0; i < fwdData->ifaceNumber; i++) {
	  forwarder_thread_t* fwdThread = &(fwdData->forwarderThread[i]);
	  fwdThread->fwdData = fwdData;
	  fwdThread->ifaceIndex = i;
	  fwdThread->running = 1;
	  fwdThread->stopped = 0;
	  if(pthread_create((void*) &(fwdThread->thread),
			    NULL,
			    forwardFunc,
			    (void*) fwdThread) != 0) {
	       LOG_ERROR("Creation of forwarding thread failed: %s\n", strerror(errno));
	       goto pthread_exit_err;
	  }

	  if (pthread_setschedprio(fwdThread->thread, (priorityMax / 3 * 2))) {
	       LOG_ERROR("Error setting forwarding thread priority (%d)\n", i);
	       goto pthread_exit_err;
	  }
     }

     LOG("Initialized %d forwarding thread(s)\n", i);

     if(pthread_create( (void*) &(fwdData->timeThread), NULL,
			timerFunc, (void*) fwdData ) != 0 ) {
	  LOG_ERROR("Creation of timer thread failed: %s\n", strerror(errno));
	  goto pthread_exit_err;
     }

     if(fwdData->pcapReader == NULL && fwdData->classifier != NULL) {
	  int tn;
	  for(tn = 0; tn < CLASS_THREAD_NUMBER; tn++) {
	       classifier_thread_t* clsThread = &(fwdData->classifierThread[tn]);
	       clsThread->fwdData = fwdData;
	       clsThread->id = tn;
	       clsThread->classifying = 1;
	       clsThread->stopped = 0;
	       if(pthread_create( (void*) &(clsThread->thread),
				  NULL,
				  classFunc,
				  (void*) clsThread ) != 0 ) {
		    LOG_ERROR("Creation of classifier thread failed: %s\n", strerror(errno));
		    goto pthread_exit_err;
	       }
	  
	       LOG("Priority classification thread %d\n", priorityMin);
	  
	       if(pthread_setschedprio(clsThread->thread, priorityMin)) {
		    LOG_ERROR("Error setting classification thread priority to %d\n", priorityMin);
		    goto pthread_exit_err;
	       }
	  }
     }

     if(fwdData->pcapReader != NULL) {
	  forwarder_thread_t* pcapThread = &(fwdData->pcapThread);
	  pcapThread->fwdData = fwdData;
	  pcapThread->ifaceIndex = -1;
	  pcapThread->running = 1;
	  pcapThread->stopped = 0;
	  if(pthread_create((void*) &(pcapThread->thread),
			    NULL,
			    pcapFunc,
			    (void*) pcapThread) != 0) {
	       LOG_ERROR("Creation of pcap reader thread failed: %s\n", strerror(errno));
	       goto pthread_exit_err;
	  }
	  LOG("Initialized pcap reader thread\n");
     }

     return 1;

pthread_exit_err:
     return 0;
}

int main_loop(forwarder_t* fwdData) {
     LOG_DEBUG("Sleeping\n");
     int n, receivedPackets;
     struct timeval ts;
     int delay = TERMINATE_DELAY;
     int totOldFlows = 0;
     int totAllocFlows = 0;

     if(sleep(2) > 0) {
	  char string[1000];
	  gettimeofday(&ts, NULL);
	  delay =
	       (ts.tv_sec - fwdData->interruptTime.tv_sec) * 1000000 +
	       (ts.tv_usec - fwdData->interruptTime.tv_usec);
	  memcpy(&(fwdData->interruptTime), &ts, sizeof(struct timeval));

	  if(fwdData->pcapReader != NULL && fwdData->pcapThread.running == 0)
	       goto skip_display;

	  if(delay < TERMINATE_DELAY)
	       goto skip_display;

	  if(fwdData->ifaceNumber > 0) {
	       string[0] = 0x0;
	       for(n = 0; n < fwdData->ifaceNumber; n++)
		    sprintf(string + strlen(string), "%d ",
			    fwdData->forwarderThread[n].receivedPackets);
	       LOG("Received packets = [%s] ", string);
	       string[0] = 0x0;
	       for(n = 0; n < fwdData->ifaceNumber; n++)
		    sprintf(string + strlen(string), "%d ",
			    fwdData->forwarderThread[n].inputErrors);
	       LOG("Input errors = [%s] ", string);
	       string[0] = 0x0;
	       for(n = 0; n < fwdData->ifaceNumber; n++)
		    sprintf(string + strlen(string), "%d/%d ",
			    fwdData->forwarderThread[n].outputErrors,
			    fwdData->forwarderThread[n].outputFailed);
	       LOG("Output errors = [%s]\n", string);
	       string[0] = 0x0;
	       for(n = 0; n < fwdData->ifaceNumber; n++) {
		    totOldFlows += fwdData->forwarderThread[n].oldFlows;
		    sprintf(string + strlen(string), "%d ", fwdData->forwarderThread[n].oldFlows);
	       }
	       LOG("Old flows = [%s] [%d]\n", string, totOldFlows);
	       string[0] = 0x0;
	       for(n = 0; n < fwdData->ifaceNumber; n++) {
		    totAllocFlows += fwdData->forwarderThread[n].totAllocFlows;
		    sprintf(string + strlen(string), "%d ",
			    fwdData->forwarderThread[n].totAllocFlows);
	       }
	       LOG("Tot alloc flows = [%s] [%d]\n", string, totAllocFlows);
	  } else {
	       LOG("Read packets = [%d]\n", fwdData->pcapThread.receivedPackets);
	       LOG("Old flows = [%d]\n", fwdData->pcapThread.oldFlows);
	       LOG("Tot alloc flows = [%d]\n", fwdData->pcapThread.totAllocFlows);
	  }
     }

     for(n = 0, receivedPackets = 0; n < fwdData->ifaceNumber; n++)
	  receivedPackets += fwdData->forwarderThread[n].receivedPackets;
     receivedPackets += fwdData->pcapThread.receivedPackets;

     LOG( "Time = %d, Packets: %d collected, %d processed (of them %d were non IP45), "
	  "%d classified, %d not supported, "
	  "%d not classified, jobRing %d, seen flows %d\n",
	  fwdData->currentTimeSec - fwdData->startTimeSec,
	  receivedPackets, fwdData->processedPackets, fwdData->nonIp45Packets,
	  fwdData->classifiedPackets, fwdData->notSupportedPackets,
	  fwdData->notClassifiedPackets,
	  fwdData->nextJobToStore - fwdData->nextJobToProcess,
	  fwdData->flowNumber);

skip_display:     
     // check if terminated when reading a pcap
     if(fwdData->pcapReader != NULL && fwdData->pcapThread.stopped == 1) {
	  return 0;
     }

     // check if terminated by user
     if(delay < TERMINATE_DELAY) return 0;
     return 1;
}

void usage() {
     char* usage_str = "sniff version 1.0.0\n"
          "     Usage: sniff [h] [ -i input_interface,output_interface ]\n"
	  "                   [ -r pcap_trace_name ]\n"
          "                   [ -c classifier_library.so ]\n"
	  "                   [ -m modelsDirectory ]\n"
	  "\n"
	  "          -i    if output_interface is not given, no packet is forwarded\n"
	  "          -c    points to the classification library module that implements\n"
	  "                the classification algorithm\n"
	  "          -m    points to the directory containing the statistical models\n"
	  "                for the specified algorithm\n"
	  "          -r    read packets from trace file instead of interface\n"
	  "          -h    prints this message\n";
     LOG("%s\n", usage_str);
}

void killThreads(forwarder_t* fwdData) {

     int n;

     if(fwdData->pcapReader == NULL && fwdData->classifier) {
	  for(n = 0; n < CLASS_THREAD_NUMBER; n++) {
	       classifier_thread_t* clsThread = &(fwdData->classifierThread[n]);
	       LOG_DEBUG("Asking classifier to terminate\n");
	       clsThread->classifying = 0;
	  }
	  for(n = 0; n < CLASS_THREAD_NUMBER; n++) {
	       classifier_thread_t* clsThread = &(fwdData->classifierThread[n]);
	       LOG_DEBUG("Waiting for classifier to terminate\n");
	       while(clsThread->stopped == 0) {};
	       LOG_DEBUG("Classifier %d stopped\n", n);
	  }
     }

     for(n = 0; n < fwdData->ifaceNumber; n++) {
	  forwarder_thread_t* fwdThread = &(fwdData->forwarderThread[n]);
	  LOG_DEBUG("Waiting for forwarder %d to terminate\n", n);
	  fwdThread->running = 0;
	  // interrupt recvfrom if blocked
	  pthread_kill(fwdThread->thread, SIGINT);
	  while(fwdThread->stopped == 0) {};
	  LOG_DEBUG("Forwarder %d stopped\n", n);
     }

     if(fwdData->pcapReader != NULL) {
	  forwarder_thread_t* pcapThread = &(fwdData->pcapThread);
	  LOG_DEBUG("Waiting pcap reader to terminate\n");
	  pcapThread->running = 0;
	  while(pcapThread->stopped == 0) {};
	  LOG_DEBUG("Pcap reader stopped\n");
     }
	  
     // CODE MISSING TO STOP THE OTHER THREADS (? TIMERFUNC)
}

void display_stats(forwarder_t* fwdData) {
     int i, k, tmp_flow = 0;
     int allocatedFlows[10] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0,};
     int usedHashRows = 0;
     int unusedHashRows = 0;
     int totNotEmptyItems = 0;
     int totItems = 0;
     int maxLength = 0;
     int oldFlowFinal = 0;
     int receivedPackets = 0;

     if (fwdData->classifier != NULL)
          fwdData->classifier->stats_fn();

     for(i = 0; i < FLOW_TABLE_SIZE; i++) {
	  flow_t* flow;
	  flow = fwdData->flowTable[i].first;
	  if(flow->empty == 0) {
	       usedHashRows++;

	       tmp_flow = -1;

	       for(flow = fwdData->flowTable[i].first;
		   flow != NULL;
		   flow = flow->next ) {

		    if(flow->empty == 0 &&
		       (fwdData->lastPacketTimeSec - flow->timeLastSeen > MAX_FLOW_AGE_SEC))
			 oldFlowFinal++;

		    totItems++;
		    if(flow->empty == 0) {
			 totNotEmptyItems++;
		    }

		    if(flow->next != NULL)
			 tmp_flow++;
	       }

	       if (tmp_flow > maxLength )
		    maxLength = tmp_flow;

	       if(tmp_flow != -1) {
		    int idx = (tmp_flow);
		    if (idx > 9) 
			 allocatedFlows[9] += idx;
		    else
			 allocatedFlows[idx]++;	
	       }
	  }
	  else
	       unusedHashRows++;
     }

     LOG("Size of flow: %ld\n", sizeof(flow_t));
     LOG("Used rows in hash table: %d\n", usedHashRows);
     LOG("Unused rows in hash table: %d\n", unusedHashRows);
     LOG("Tot elements (empty and not): %d\n",totItems);
     LOG("Tot not empty elements: %d\n",totNotEmptyItems);
     LOG("Tot old elements in list not yet marked: %d\n",oldFlowFinal);
     LOG("Longest list size: %d\n",maxLength+1);

     for(k = 0; k < 10; k++) {
	  if(k == 9)
	       LOG("Lists in hash table with %d elements or more: %d\n", k + 1, allocatedFlows[k]);
	  else
	       LOG("Lists in hash table with exactly %d elements: %d\n", k + 1, allocatedFlows[k]);
     }

#ifdef LOG_FLOW
     char string[LOG_FLOW * 30];
     int total_classified = 0;
     string[0] = 0x0;
     LOG("flows whose Nth packet was marked\n");
     for(k = 0; k < LOG_FLOW; k++) {
	  sprintf(string + strlen(string), "%d ", fwdData->flow_whose_Nth_packet_was_marked[k]);
	  total_classified += fwdData->flow_whose_Nth_packet_was_marked[k];
     }
     LOG("%s\n", string);
     string[0] = 0x0;
     LOG("flows classified exactly after N packets\n");
     for(k = 0; k < LOG_FLOW; k++) {
	  sprintf(string + strlen(string), "%d ", fwdData->flow_classified_exactly_after_N_packets[k]);
	  total_classified += fwdData->flow_classified_exactly_after_N_packets[k];
     }
     LOG("%s\n", string);
     string[0] = 0x0;
     LOG("flows with at least N packets\n");
     for(k = 0; k < LOG_FLOW; k++) {
	  sprintf(string + strlen(string), "%d ", fwdData->flow_with_at_least_N_packets[k]);
	  total_classified += fwdData->flow_with_at_least_N_packets[k];
     }
     LOG("%s\n", string);
     LOG("classified flows: %d\nseen flows: %d\n", fwdData->classified_flows, fwdData->flowNumber);
#endif // LOG_FLOW 

     for(k = 0, receivedPackets = 0; k < fwdData->ifaceNumber; k++)
	  receivedPackets += fwdData->forwarderThread[k].receivedPackets;
     receivedPackets += fwdData->pcapThread.receivedPackets;
     LOG("Received packets: %d\n", receivedPackets);
     LOG("Non IP 45 packets: %d\n", fwdData->nonIp45Packets);
     LOG("Non IP packets: %d\n", fwdData->nonIpPackets);

#ifdef LOG_FLOW_ENDTIME
     for(k = 0; k < FLOW_TABLE_SIZE; k++) {
#define FLOW_STRING_LENGTH 100
	  char flow_string[FLOW_STRING_LENGTH];
	  assert(PACKETS_TO_WAIT >= 4);
          flow_t* flow;
          for(flow = fwdData->flowTable[k].first; flow != NULL; flow = flow->next ) {
               if(flow->empty == 0) {
                    struct in_addr* ipsrc = (struct in_addr*) &(flow->ipsrc);
                    struct in_addr* ipdst = (struct in_addr*) &(flow->ipdst);
		    snprintf(flow_string,
			     FLOW_STRING_LENGTH,
			     "%s:%hu -> ", inet_ntoa(*ipsrc), ntohs(flow->portsrc));
		    snprintf(flow_string + strlen(flow_string),
			     FLOW_STRING_LENGTH - strlen(flow_string),
			     "%s:%hu (%hhu) %d",
			     inet_ntoa(*ipdst), ntohs(flow->portdst),
			     flow->proto, flow->seenPackets);
		    snprintf(flow_string + strlen(flow_string),
			     FLOW_STRING_LENGTH - strlen(flow_string),
			     " %d %d %d %d",
			     flow->packet_sizes[0], flow->packet_sizes[1], 
			     flow->packet_sizes[2], flow->packet_sizes[3]);
		    snprintf(flow_string + strlen(flow_string),
			     FLOW_STRING_LENGTH - strlen(flow_string),
			     " %d", flow->classificationCode);
		    LOG("show flow - %s\n", flow_string);
               }
          }
     }
#endif // LOG_FLOW_ENDTIME
}

int main(int argc, char* argv[])
{
     forwarder_t fwdData;
     char* className = NULL;
     char* classDir = NULL;
     int c;

     assert(sizeof(muxhdr_t) == 8);

     memset(&fwdData, 0, sizeof(forwarder_t));

     while((c = getopt(argc, argv, "r:i:c:m:h")) != -1) {
	  switch(c) {
	  case 'c':
	       className = optarg;
	       break;
	  case 'm':
	       classDir = optarg;
	       break;
	  case 'i':
               if(fwdData.ifaceNumber == MAX_IFACE) {
		    LOG("Too many interfaces requested, try increasing MAX_IFACE\n");
                    goto exit_error;
               }
	       fwdData.interfaceName[fwdData.ifaceNumber] = optarg;
               fwdData.ifaceNumber++;
	       break;
	  case 'r':
	       fwdData.traceName = optarg;
	       break;
	  case 'h':
	       usage();
	  default:
	       goto exit_error;
	  }
     }

     if(fwdData.ifaceNumber == 0 && fwdData.traceName == NULL) {
	  LOG_ERROR("No interface given nor trace name\n");
          goto exit_error;
     }

     if(fwdData.ifaceNumber > 0 && fwdData.traceName != NULL) {
	  LOG_ERROR("Cannot set interface and trace name at the same time\n");
	  goto exit_error;
     }

     if(initMainLoop(&fwdData) == 0)
	  goto exit_error;

     if(initFlowTable(&fwdData, FLOW_TABLE_SIZE) == 0)
	  goto exit_error;

     if(initClassifier(&fwdData, JOB_RING_SIZE, className, classDir) == 0)
	  goto exit_error;

     if(fwdData.ifaceNumber > 0 && initForwarder(&fwdData) == 0)
	  goto exit_error;

     if(fwdData.traceName != NULL && initPcapReader(&fwdData) == 0)
	  goto exit_error;

     if(initThread(&fwdData) == 0)
	  goto exit_error;

     while(main_loop(&fwdData)) {};

     killThreads(&fwdData);

     display_stats(&fwdData);

     return 0;

exit_error:
     return -1;
}
